"""Generic pwm driver."""
from __future__ import division
import math


class Driver:
    """Represents the base class for pwm drivers."""

    IO_TRIES = 10

    def __init__(self, pins, resolution, freq):
        """
        Initialize the driver.

        :param pins: The pin numbers that should be controlled.
        :param resolution: The resolution of the pwm channel.
        :param freq: The pwm frequency.
        """
        if not isinstance(pins, list):
            pins = [pins]

        self._pins = pins
        self._resolution = resolution
        self._freq = freq
        self._state = [0] * len(self._pins)
        self._max_raw_value = math.pow(2, self._resolution) - 1

    @property
    def pins(self):
        """
        Color property.

        :return: The pins numbers that are controlled.
        """
        return self._pins

    def set_pwm(self, values):
        """
        Set pwm values on the controlled pins.

        :param values: Values to set (0.0-1.0).
        :return:
        """
        if len(values) != len(self._pins):
            raise ValueError('Number of values has to be identical with '
                             'the number of pins.')
        if not all(0 <= v <= 1 for v in values):
            raise ValueError('Values must be between 0 and 1.')

        for tries in range(self.IO_TRIES):
            try:
                self._set_pwm(self._to_raw_pwm(values))
                break
            except IOError as error:
                if tries == self.IO_TRIES - 1:
                    raise error
        self._state = values

    def _set_pwm(self, values):
        """
        Method stub for setting the pwm values.

        Has to be implemented by inheriting classes.
        :param values: Values to set (0.0-1.0).
        """
        raise NotImplementedError

    def _to_raw_pwm(self, values):
        """
        Convert uniform pwm values to raw, driver-specific values.

        :param values: The uniform pwm values (0.0-1.0).
        :return: Converted, driver-specific pwm values.
        """
        return [self._to_single_raw_pwm(values[i])
                for i in range(len(self._pins))]

    def _to_single_raw_pwm(self, value):
        """
        Convert a single uiform pwm value to raw, driver-specific value.

        :param value: The uniform pwm value (0.0-1.0).
        :return: Converted, driver-specific pwm value.
        """
        return int(round(value * self._max_raw_value))

    def _to_uniform_pwm(self, values):
        """
        Convert raw pwm values to uniform values.

        :param values: The raw pwm values.
        :return: Converted, uniform pwm values (0.0-1.0).
        """
        return [self._to_single_uniform_pwm(values[i])
                for i in range(len(self._pins))]

    def _to_single_uniform_pwm(self, value):
        """
        Convert a single raw pwm value to uniform value.

        :param value: The raw pwm value.
        :return: Converted, uniform pwm value (0.0-1.0).
        """
        return value / self._max_raw_value

    def steps(self, start, end):
        """
        Get the maximum number of steps the driver needs for a transition.

        :param start: The start value as uniform pwm value (0.0-1.0).
        :param end: The end value as uniform pwm value (0.0-1.0).
        :return: The maximum number of steps.
        """
        if not 0 <= start <= 1:
            raise ValueError('Values must be between 0 and 1.')
        if not 0 <= end <= 1:
            raise ValueError('Values must be between 0 and 1.')

        raw_start = self._to_single_raw_pwm(start)
        raw_end = self._to_single_raw_pwm(end)
        return abs(raw_start - raw_end)

    def stop(self):
        """Stop the driver and release resources."""
        self._stop()

    def _stop(self):
        """
        Method stub for stopping the driver.

        Has to be implemented by inheriting classes.
        """
        pass



"""GPIO pwm driver."""
import pigpio

from pwmled.driver import Driver


class GpioDriver(Driver):
    """Represents a pwm driver, which uses GPIOs of a Raspberry Pi."""

    RESOLUTION = 8

    def __init__(self, pins, freq=200, host=None):
        """
        Initialize the driver.

        :param pins: The pin numbers, that should be controlled.
        :param freq: The pwm frequency.
        :param host: The host name of the Pi on which the pigpio daemon is running.
        """
        super().__init__(pins, self.RESOLUTION, freq)

        if host is not None:
            self._pi = pigpio.pi(host)
        else:
            self._pi = pigpio.pi()

    def _set_pwm(self, raw_values):
        """
        Set pwm values on the controlled pins.

        :param raw_values: Raw values to set (0-255).
        """
        for i in range(len(self._pins)):
            self._pi.set_PWM_dutycycle(self._pins[i], raw_values[i])

    def _stop(self):
        """Stop the driver and release resources."""
        self._pi.stop()



"""Transition of a led."""
import time
import math
import threading


class Transition:
    """Represents a transition of a led."""

    def __init__(self, driver, duration, state_stages,
                 pwm_stages, callback=None):
        """
        Initialize the transition.

        :param driver: The driver of the led.
        :param duration: The duration.
        :param state_stages: The state stages to be executed.
        :param pwm_stages: The pwm stages to be executed.
        :param callback: Callback that is called when transition has finished.
        """
        self._driver = driver
        self._callback = callback
        self.duration = duration
        self.state_stages = state_stages
        self.pwm_stages = pwm_stages

        self.stage_index = 0
        self.cancelled = False
        self.finished = False
        self._finish_event = threading.Event()
        self._start_time = time.time()

    def step(self):
        """Apply the current stage of the transition based on current time."""
        if self.cancelled or self.finished:
            return

        if not self.pwm_stages:
            self._finish()
            return

        if self.duration == 0:
            progress = 1
        else:
            run_time = time.time() - self._start_time
            progress = max(0, min(1, run_time / self.duration))
        self.stage_index = math.ceil((len(self.pwm_stages) - 1) * progress)
        stage = self.pwm_stages[self.stage_index]
        self._driver.set_pwm(stage)

        if progress == 1:
            self._finish()

    def _finish(self):
        """Mark transition as finished and execute callback."""
        self.finished = True
        if self._callback:
            self._callback(self)
        self._finish_event.set()

    def wait(self, timeout=None):
        """
        Wait for transition to be finished.

        :param timeout: Timeout of the operation in seconds.
        """
        self._finish_event.wait(timeout=timeout)

    def cancel(self):
        """
        Cancel the transition.
        """
        self._finish()
        self.cancelled = True



"""RGB led controller."""
from __future__ import division

from pwmled import Color
from pwmled.led import SimpleLed


class RgbLed(SimpleLed):
    """Represents a RGB led that can be controlled."""

    def __init__(self, driver):
        """
        Initialize the led.

        :param driver: The driver that is used to control the led.
        """
        super().__init__(driver)
        self._color = Color(255, 255, 255)

    @property
    def color(self):
        """
        The color property.

        :return: The color of the led.
        """
        return self._color

    @color.setter
    def color(self, color):
        """
        Set the color of the led.

        :param color: The color to set.
        """
        self.set(color=color)

    def set(self, is_on=None, brightness=None, color=None,
            cancel_transition=True):
        """
        Set properties of the led simultaneously before updating pwm values.

        :param is_on: On-off state of the led.
        :param brightness: Brightness of the led.
        :param color: Color of the led.
        :param cancel_transition: Cancel active transitions.
        """
        if cancel_transition:
            self._cancel_active_transition()

        if color is not None:
            self._assert_is_color(color)
            self._color = color

        super().set(is_on, brightness, cancel_transition=False)

    def _get_pwm_values(self, brightness=None, color=None):
        """
        Get the pwm values for a specific state of the led.

        If a state argument is omitted, current value is used.

        :param brightness: The brightness of the state.
        :param color: The color of the state.
        :return: The pwm values.
        """
        if brightness is None:
            brightness = self.brightness
        if color is None:
            color = self.color

        return [(x / 255) * brightness for x in color]

    def _prepare_transition(self, is_on=None, brightness=None, color=None):
        """
        Perform pre-transition tasks and construct the destination state.

        :param is_on: The on-off state to transition to.
        :param brightness: The brightness to transition to (0.0-1.0).
        :param color: The color to transition to.
        :return: The destination state of the transition.
        """
        dest_state = super()._prepare_transition(is_on,
                                                 brightness=brightness,
                                                 color=color)

        # Handle transitions from off to on and changing color
        if is_on and not self.is_on and color is not None:
            self.set(color=color, cancel_transition=False)

        return dest_state

    def _transition_steps(self, brightness=None, color=None):
        """
        Get the maximum number of steps needed for a transition.

        :param brightness: The brightness to transition to (0.0-1.0).
        :param color: The color to transition to.
        :return: The maximum number of steps.
        """
        values = []
        if brightness is not None:
            self._assert_is_brightness(brightness)
            values.append((self.brightness, brightness))

        if color is not None:
            self._assert_is_color(color)
            values += [(self.color[i] / 255, color[i] / 255) for i in range(3)]

        if not values:
            return 0

        return max(self._driver.steps(*args) for args in values)

    def _transition_stage(self, step, total_steps,
                          brightness=None, color=None):
        """
        Get a transition stage at a specific step.

        :param step: The current step.
        :param total_steps: The total number of steps.
        :param brightness: The brightness to transition to (0.0-1.0).
        :param color: The color to transition to.
        :return: The stage at the specific step.
        """
        if brightness is not None:
            self._assert_is_brightness(brightness)
            brightness = self._interpolate(self.brightness, brightness,
                                           step, total_steps)

        if color is not None:
            self._assert_is_color(color)
            color = Color(*[self._interpolate(self.color[i], color[i],
                                              step, total_steps)
                            for i in range(3)])

        return {'brightness': brightness, 'color': color}

    @staticmethod
    def _assert_is_color(value):
        """
        Assert that the given value is a valid brightness.

        :param value: The value to check.
        """
        if not isinstance(value, tuple) or len(value) != 3:
            raise ValueError("Color must be a RGB tuple.")

        if not all(0 <= x <= 255 for x in value):
            raise ValueError("RGB values of color must be between 0 and 255.")



"""RGBW led controller."""
from __future__ import division

from pwmled.led.rgb import RgbLed


class RgbwLed(RgbLed):
    """Represents a RGBW led that can be controlled."""

    def _get_pwm_values(self, brightness=None, color=None):
        """
        Get the pwm values for a specific state of the led.

        If a state argument is omitted, current value is used.

        :param brightness: The brightness of the state.
        :param color: The color of the state.
        :return: The pwm values.
        """
        if brightness is None:
            brightness = self.brightness
        if color is None:
            color = self.color

        return [(x / 255) * brightness for x in self._rgb_to_rgbw(color)]

    @staticmethod
    def _rgb_to_rgbw(color):
        """
        Convert a RGB color to a RGBW color.

        :param color: The RGB color.
        :return: The RGBW color.
        """
        # Get the maximum between R, G, and B
        max_value = max(color)

        # If the maximum value is 0, immediately return pure black.
        if max_value == 0:
            return [0] * 4

        # Figure out what the color with 100% hue is
        multiplier = 255 / max_value
        hue_red = color.R * multiplier
        hue_green = color.G * multiplier
        hue_blue = color.B * multiplier

        # Calculate the whiteness (not strictly speaking luminance)
        max_hue = max(hue_red, hue_green, hue_blue)
        min_hue = min(hue_red, hue_green, hue_blue)
        luminance = ((max_hue + min_hue) / 2 - 127.5) * 2 / multiplier

        # Calculate the output values
        r = max(0, min(255, int(color.R - luminance)))
        g = max(0, min(255, int(color.G - luminance)))
        b = max(0, min(255, int(color.B - luminance)))
        w = max(0, min(255, int(luminance)))

        return [r, g, b, w]



"""Manager for led transitions."""
import time
import threading
from singleton import Singleton


class TransitionManager(object, metaclass=Singleton):
    """Represents a manager that executes transitions in a separate thread."""

    MIN_STEP_TIME = 0.001

    def __init__(self):
        """Initialize the manager."""
        self._thread = None
        self._transitions = []

    def execute(self, transition):
        """
        Queue a transition for execution.

        :param transition: The transition
        """
        self._transitions.append(transition)
        if self._thread is None or not self._thread.is_alive():
            self._thread = threading.Thread(target=self._transition_loop, daemon=True)
            self._thread.start()

    def _transition_loop(self):
        """Execute all queued transitions step by step."""
        while self._transitions:
            start = time.time()
            for transition in self._transitions:
                transition.step()
                if transition.finished:
                    self._transitions.remove(transition)

            time_delta = time.time() - start
            sleep_time = max(0, self.MIN_STEP_TIME - time_delta)
            time.sleep(sleep_time)



"""Simple led controller."""

from functools import partial

from pwmled.transitions.transition import Transition
from pwmled.transitions.transition_manager import TransitionManager


class SimpleLed:
    """Represents a simple, one-color led that can be controlled."""

    def __init__(self, driver):
        """
        Initialize the led.

        :param driver: The driver that is used to control the led.
        """
        self._driver = driver
        self._is_on = False
        self._brightness = 1
        self._active_transition = None

    @property
    def is_on(self):
        """
        On-off state property.

        :return: True, if led is on. False otherwise.
        """
        return self._is_on

    def on(self):
        """Turn the led on."""
        self.set(is_on=True)

    def off(self):
        """Turn the led off."""
        self.set(is_on=False)

    @property
    def brightness(self):
        """
        Brightness property.

        :return: The brightness of the led (0.0-1.0).
        """
        return self._brightness

    @brightness.setter
    def brightness(self, brightness):
        """
        Set the brightness of the led.

        :param brightness: The brightness to set (0.0-1.0).
        """
        self.set(brightness=brightness)

    def set(self, is_on=None, brightness=None, cancel_transition=True):
        """
        Set properties of the led simultaneously before updating pwm values.

        :param is_on: On-off state of the led.
        :param brightness: Brightness of the led.
        :param cancel_transition: Cancel active transitions.
        """
        if cancel_transition:
            self._cancel_active_transition()

        if is_on is not None:
            self._is_on = is_on

        if brightness is not None:
            self._assert_is_brightness(brightness)
            self._brightness = brightness

        self._update_pwm()

    def _update_pwm(self):
        """Update the pwm values of the driver regarding the current state."""
        if self._is_on:
            values = self._get_pwm_values()
        else:
            values = [0] * len(self._driver.pins)

        self._driver.set_pwm(values)

    def _get_pwm_values(self, brightness=None):
        """
        Get the pwm values for a specific state of the led.

        If a state argument is omitted, current value is used.

        :param brightness: The brightness of the state.
        :return: The pwm values.
        """
        if brightness is None:
            brightness = self.brightness

        return [brightness]

    def transition(self, duration, is_on=None, **kwargs):
        """
        Transition to the specified state of the led.

        If another transition is already running, it is aborted.

        :param duration: The duration of the transition.
        :param is_on: The on-off state to transition to.
        :param kwargs: The state to transition to.
        """
        self._cancel_active_transition()

        dest_state = self._prepare_transition(is_on, **kwargs)
        total_steps = self._transition_steps(**dest_state)
        state_stages = [self._transition_stage(step, total_steps, **dest_state)
                        for step in range(total_steps)]
        pwm_stages = [self._get_pwm_values(**stage)
                      for stage in state_stages]
        callback = partial(self._transition_callback, is_on)
        self._active_transition = Transition(self._driver, duration,
                                             state_stages, pwm_stages,
                                             callback)

        TransitionManager().execute(self._active_transition)

        return self._active_transition

    def _cancel_active_transition(self):
        if self._active_transition:
            self._active_transition.cancel()
            self._active_transition = None

    def _transition_callback(self, is_on, transition):
        """
        Callback that is called when a transition has ended.

        :param is_on: The on-off state to transition to.
        :param transition: The transition that has ended.
        """
        # Update state properties
        if transition.state_stages:
            state = transition.state_stages[transition.stage_index]
            if self.is_on and is_on is False:
                # If led was turned off, set brightness to initial value
                # so that the brightness is restored when it is turned on again
                state['brightness'] = self.brightness
            self.set(is_on=is_on, cancel_transition=False, **state)

        self._active_transition = None

    def _prepare_transition(self, is_on=None, **kwargs):
        """
        Perform pre-transition tasks and construct the destination state.

        :param is_on: The on-off state to transition to.
        :param kwargs: The state to transition to.
        :return: The destination state of the transition.
        """
        # Handle transitions with changing on-off-state
        dest_state = kwargs.copy()
        if is_on is not None:
            if is_on and not self.is_on:
                if 'brightness' not in kwargs or kwargs['brightness'] is None:
                    dest_state['brightness'] = self.brightness
                self.set(brightness=0, cancel_transition=False)
            elif not is_on and self.is_on:
                dest_state['brightness'] = 0

        return dest_state

    def _transition_steps(self, brightness=None):
        """
        Get the maximum number of steps needed for a transition.

        :param brightness: The brightness to transition to (0.0-1.0).
        :return: The maximum number of steps.
        """
        if brightness is not None:
            self._assert_is_brightness(brightness)
            return self._driver.steps(self.brightness, brightness)

        return 0

    def _transition_stage(self, step, total_steps, brightness=None):
        """
        Get a transition stage at a specific step.

        :param step: The current step.
        :param total_steps: The total number of steps.
        :param brightness: The brightness to transition to (0.0-1.0).
        :return: The stage at the specific step.
        """
        if brightness is not None:
            self._assert_is_brightness(brightness)
            brightness = self._interpolate(self.brightness, brightness,
                                           step, total_steps)

        return {'brightness': brightness}

    @staticmethod
    def _interpolate(start, end, step, total_steps):
        """
        Interpolate a value from start to end at a given progress.

        :param start: The start value.
        :param end: The end value.
        :param step: The current step.
        :param total_steps: The total number of steps.
        :return: The interpolated value at the given progress.
        """
        diff = end - start
        progress = step / (total_steps - 1)
        return start + progress * diff

    @staticmethod
    def _assert_is_brightness(value):
        """
        Assert that the given value is a valid brightness.

        :param value: The value to check.
        """
        if not 0 <= value <= 1:
            raise ValueError('Brightness must be between 0 and 1.')



"""PCA9685 pwm driver."""
import board
import busio
import adafruit_pca9685

from pwmled.driver import Driver


class Pca9685Driver(Driver):
    """Represents a pwm driver, which uses the pins of an PCA9685 I2C board."""

    RESOLUTION = 16

    def __init__(self, pins, freq=200, address=0x40):
        """
        Initialize the driver.

        :param pins: The pin numbers, that should be controlled.
        :param freq: The pwm frequency.
        :param address: The address of the PCA9685.
        """
        super().__init__(pins, self.RESOLUTION, freq)

        i2c = busio.I2C(board.SCL, board.SDA)
        self._device = adafruit_pca9685.PCA9685(i2c, address=address)
        self._device.frequency = self._freq

    def _set_pwm(self, raw_values):
        """
        Set pwm values on the controlled pins.

        :param raw_values: Raw values to set (0-65535).
        """
        for i in range(len(self._pins)):
            channel = self._device.channels[self._pins[i]]
            channel.duty_cycle = raw_values[i]

    def _stop(self):
        """Stop the driver and release resources."""
        self._device.deinit()



from setuptools import setup, find_packages


with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='pwmled',
    version='1.5.3',
    description='Control LEDs connected to a micro controller using pwm.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/soldag/python-pwmled/',
    license='MIT',
    author='soldag',
    author_email='soren.oldag@gmail.com',
    packages=find_packages(),
    install_requires=[
        'pigpio==1.35',
        'adafruit-blinka==5.5.1',
        'adafruit-circuitpython-pca9685==3.3.2',
        'python-singleton==0.1.2',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ]
)
